# Plugin that creates an STL file representing the current model
# Author: K. Di Filippo

# Global requires
require 'sketchup.rb'

# Load dependencies
PLUGIN_PATH = 'Virtualwind/'
Sketchup.load PLUGIN_PATH + 'vwGlobalConstants'
Sketchup.load PLUGIN_PATH + 'textureScripts'



# Top-level function to toggle the "Export With Textures" option.  
# All this tool does is toggle the exportWithTextures flag in the
# VW_GLOBAL_DATA_CONTAINER; the checked/unchecked state of the Tool
# is handled by the set_validation_proc feature of the Tool's 
# underlying Command.
def toggleExportWithTextures

  VW_GLOBAL_DATA_CONTAINER.exportWithTextures = !VW_GLOBAL_DATA_CONTAINER.exportWithTextures?

end




# Top-level function to export the active model to a VWS file.
# Params:
#   - export_type: the "filter" used to determine what will be considered for 
#     export and how.  export_type is compared with the EXPORT_MASK_[...]
#     constants, and so should conform to the EXPORT_MASK conventions.  For
#     example, if EXPORT_MASK_SHEETS is specified, only Sheet geometry
#     will be exported.
def export_vws_file export_type
  
  # Create and start the Progress Bar.
  progress_bar = vwProgressBar.new
  progress_bar.start

  # Decode the export type.
  export_force_solids = (export_type & EXPORT_MASK_FORCE_SOLIDS) > 0
  export_force_sheets = (export_type & EXPORT_MASK_FORCE_SHEETS) > 0
  export_force_terrain = (export_type & EXPORT_MASK_FORCE_TERRAIN) > 0
  export_force_none = !(export_force_solids || export_force_sheets || export_force_terrain )
  export_solids = (export_type & EXPORT_MASK_SOLIDS) > 0
  export_sheets = (export_type & EXPORT_MASK_SHEETS) > 0
  export_terrain = (export_type & EXPORT_MASK_TERRAIN) > 0
  
  # Determine if Textures are to exported.  (Just check the flag in 
  # the Global Data Container and store it locally for clarity.)
  export_textures = VW_GLOBAL_DATA_CONTAINER.exportWithTextures?

  # Bring up the Save dialog box (using the model's filename as the 
  # default filename for the VWS file, or "untitled" if no Model is 
  # loaded.).
  default_name = File.basename(Sketchup.active_model.path, '.*')
  default_name = 'untitled' if default_name == ''
  default_extension = '.vws'
  save_file_name_and_path = UI.savepanel('Specify output file', '',  default_name + default_extension)

  # Ensure that the user didn't hit "Cancel" nor entered a blank filename.
  if (save_file_name_and_path == nil) || (save_file_name_and_path == '')
    progress_bar.clear_display
    return false
  end
  
  # Validate the filename/pathname combination.  (i.e. the file's parent
  # folder name must match the file's name.)
  save_path = File.dirname(save_file_name_and_path)
  save_name = File.basename(save_file_name_and_path, '.vws')
  if(File.basename(save_path) != save_name)
    message = "The VWS file must be saved in a folder of the same name.  Press 'Ok' to continue and have this folder automatically created, or press 'Cancel' to abort."
    answer = UI.messagebox message, MB_OKCANCEL
    if(answer == 2)
      progress_bar.clear_display
      return false
    else
      save_path = File.join(save_path, save_name)
      if File.exist?(save_path)
        Dir.chdir(save_path)
      else
        Dir.mkdir(save_path)
      end
    end
  end

  # Create the actual file, plus all necessarily sub-directories.
  save_file_name_and_path = File.join(save_path, save_name+'.vws')
  file = File.new(save_file_name_and_path, 'w')
  texture_path = File.join(save_path, 'Textures')
  if !File.exist?(texture_path)
    Dir.mkdir(texture_path)
  end
  raw_path = File.join(save_path, 'Raw')
  if !File.exist?(raw_path)
    Dir.mkdir(raw_path)
  end
  assembled_path = File.join(save_path, 'Assembled')
  if !File.exist?(assembled_path)
    Dir.mkdir(assembled_path)
  end
  
  
  # Create the VWSWriter object
  vws_writer = VWSWriter.new

  # Initialise the solid, sheet, and unknown stacks.
  solid_stack = []
  sheet_stack = []
  terrain_stack = []
  unknown_stack = []

  # Separate each top-level complex Entity into "Solid", "Sheet", and "Unknown", according to the layer the
  # Entity is on.  (Only Entities that belong to a top-level Entity are considered.)
  entities = Sketchup.active_model.entities
  entities.each{|entity|
    if entity.is_complex_entity?
      is_building = entity.layer.name == LAYER_NAME_SOLIDS
      is_canopy = entity.layer.name == LAYER_NAME_SHEETS
      is_terrain = entity.layer.name == LAYER_NAME_TERRAIN
      is_unknown = !(is_building || is_canopy || is_terrain)
      solid_stack.push entity if((is_building && export_solids && export_force_none) || export_force_solids)
      sheet_stack.push entity if((is_canopy && export_sheets && export_force_none) || export_force_sheets)
      terrain_stack.push entity if((is_terrain && export_terrain && export_force_none) || export_force_terrain)
      unknown_stack.push entity if(is_unknown && export_force_none)
    end
  }

  # Validate everything on the solid stack.
  solid_stack.each{|solid|
    separated_edges = separate_valid_and_invalid_edges solid.definition.entities
    if !separated_edges[1].empty?
      selection = Sketchup.active_model.selection
      selection.clear
      selection.add solid
      UI.messagebox "Not all geometry to be exported as a solid is valid.  Please validate the highlighted geometry or export it as a Canopy."
      progress_bar.clear_display
      return false
    end
  }


  # Reset the Progress Bar now that we have an idea of the length of the operation.
  progress_bar_length = 0
  progress_bar_length += solid_stack.length if export_solids
  progress_bar_length += sheet_stack.length if export_sheets
  progress_bar.reset(progress_bar_length)

  # If the Export Textures flag is on, export the Textures to file, 
  # and save the TextureWriter (the TW will be used later on); if 
  # off, set the texture_writer to nil (which will be interpreted as 
  # a flag later down the calling chain).
  if export_textures
    texture_writer = write_all_textures_in_model texture_path
  else
    texture_writer = nil
  end
    

  # Write the header.
  file.write vws_writer.generate_open_tag('vwsFile', nil)
  
  # write out each vwBuilding.
  if export_solids && (export_force_none || export_force_solids)
    solid_stack.each{|building|
      
      # Update the Progress Bar.
      progress_bar.increment
      
      # Get the Entity's name and its transformation matrix.
      building_name = building.name
      t_matrix = Geom::Transformation.scaling SCALE_INCHES_TO_METRES
      t_matrix = t_matrix * building.transformation
      
      # Get the Definition and write the <vwBuilding> element.
      # (Get the contents of <vwBuilding> first, then call the VWSWriter to write the <vwBuilding> element itself).
      building = building.definition
      building_element_string = ''
      building_element_string += vws_writer.generate_element('object_name', 'YES', building_name)
      building_element_string += generate_vwComponent_element(VW_COMPONENT_TYPE_BUILDING, building.entities, 'TopLevelFaces', t_matrix, texture_writer)
      file.write vws_writer.generate_element('vwBuilding', nil, building_element_string)
      
    }
  end
  
  # Write out each vwCanopy.
  if export_sheets && (export_force_none || export_force_sheets)
    sheet_stack.each{|canopy|
      
      # Update the Progress Bar.
      progress_bar.increment
      
      # Get the Entity's name and its transformation matrix.
      canopy_name = canopy.name
      t_matrix = Geom::Transformation.scaling SCALE_INCHES_TO_METRES
      t_matrix = t_matrix * canopy.transformation
      
      # Get the Definition and write the <vwCanopy> element.
      # (Get the contents of <vwCanopy> first, then call the VWSWriter to write the <vwCanopy> element itself).
      canopy = canopy.definition
      canopy_element_string = ''
      canopy_element_string += vws_writer.generate_element('object_name', 'YES', canopy_name)
      canopy_element_string += generate_vwComponent_element(VW_COMPONENT_TYPE_CANOPY, canopy.entities, 'TopLevelFaces', t_matrix, texture_writer)
      file.write vws_writer.generate_element('vwCanopy', nil, canopy_element_string)
      
    }
  end

  # Write out each vwTerrain.
  if export_terrain && (export_force_none || export_force_terrain)
    terrain_stack.each{|terrain|
      
      # Update the Progress Bar.
      progress_bar.increment
      
      # Get the Entity's name and its transformation matrix.
      terrain_name = terrain.name
      t_matrix = Geom::Transformation.scaling SCALE_INCHES_TO_METRES
      t_matrix = t_matrix * terrain.transformation
      
      # Get the Definition and write the <vwCanopy> element.
      # (Get the contents of <vwCanopy> first, then call the VWSWriter to write the <vwCanopy> element itself).
      terrain = terrain.definition
      terrain_element_string = ''
      terrain_element_string += vws_writer.generate_element('object_name', 'YES', terrain_name)
      terrain_element_string += generate_vwComponent_element(VW_COMPONENT_TYPE_TERRAIN, terrain.entities, 'TopLevelFaces', t_matrix, texture_writer)
      file.write vws_writer.generate_element('vwTerrain', nil, terrain_element_string)
      
    }
  end


  # Write the footer (close the <vwsFile> tag)
  file.write vws_writer.generate_close_tag
  
  # Close the file, tell the user the export is finished, and clear the Progress Bar when the user hits ok.
  file.close
  UI.messagebox 'VWS file export complete'
  progress_bar.clear_display

  # Set the "Last Exported File" variable in the Global Data 
  # Container, and return.
  VW_GLOBAL_DATA_CONTAINER.lastExportedFile = save_file_name_and_path
  return save_file_name_and_path
  
end



# Helper function to generate a VWS "vw[Type]Part" string (VWS-format)
# Params:
#     - vw_component_type:  The type of the vw[Type]Part being generated.  Should be one of
#       the VW_COMPONENT_TYPE_[...] constants defined in vwGlobalConstants.
#     - entities:   The array of Entity objects to process.   Can also be a Sketchup::Entities object.  
#     - component_name:  The name of the vw[Type]Part being written.  If none, pass "" or nil.
#     - t_matrix:   The Transformation object that will be applied to each Entity in entities.
#       (pass identity if no transformation is desired).
#     - texture_writer:  The TextureWriter in which the Textures have been loaded.  A value of nil
#	indicates that Textures should not be exported.
# Returns:
#     The [Type]Part string corresponding to the Entities in the entities parameter.
# Comments: 
#     - See VWS file-format documentation for details on the "vw[Type]Part" concept.
#     - In the implementation of this function, the term "Component" is often used where "Part"
#       might otherwise have been expected (in the name of the function, for example).  This is 
#       due to historical reasons.  This taxonomical quirk may be updated in future revisions.
def generate_vwComponent_element(vw_component_type, entities, component_name, t_matrix, texture_writer)
  
  # SETUP AND PREPROCESSING
  
  # Create the VWSWriter and the MeshWriter.
  vws_writer = VWSWriter.new
  
  # Split out the complex Entities from the simple Entities
  complex_entities = []
  faces_textured = []
  faces_untextured = []
  entities.each{|entity|
    
    # For complex Entity elements, just push them onto the complex_entities stack.
    if entity.is_complex_entity?
      complex_entities.push entity
    
    # For Face elements, refine the split according to whether the 
    # Face has a Texture or not (and store its "Texture Writer 
    # Handler" if it does).  (If the texture_writer is nil, label
    # everything as unTextured.)
    elsif entity.kind_of? Sketchup::Face
      material = entity.material
      if((texture_writer == nil) || (material == nil))
        faces_untextured.push entity
      else
        if material.texture == nil
          faces_untextured.push entity
        else
          uvHelper = entity.get_UVHelper(true, true, texture_writer)
          handle = texture_writer.handle(entity, true)
          faces_textured.push [entity, handle]
        end
      end
    end
  }
  
  # Once Entities have been split, sort the Faces-with-Textures array, 
  # according to the <=> function for Materials.  (faces_textured needs to
  # be sorted in order for the Face Mesh Bundles to be output correctly.)
  # Note: this can't be done with Textures (which would be more direct)
  # since the <=> is not defined for Textures (only Materials).
  faces_textured.sort! {|x,y| x[1] <=> y[1]}
  
  
  
  
  # RECURSE DOWN FOR COMPLEX ENTITIES
  
  # Declare an intermediate string
  nested_components_string = ''
  
  # For each complex Entity, recurse down.
  complex_entities.each{|entity| 
    
    # Update the transformation matrix and capture the Entity's name.
    t_matrix_new = t_matrix * entity.transformation
    entity_name = entity.name
    
    # Get the Definition and recurse down.
    entity = entity.definition
    nested_components_string += generate_vwComponent_element(vw_component_type, entity.entities, entity_name, t_matrix_new, texture_writer) 
    
  }
  
  
  

  # GENERATE THE ELEMENT STRINGS
  
  # <vwModel> element
  model_untexturedFaces_faceMesh_string = ''
  model_texturedFaces_faceMesh_string = ''

  # Get the <vwFaceMesh> strings for the blank Faces (just call write_faceMesh).
  if !faces_untextured.empty?
    model_untexturedFaces_faceMesh_string = generate_vwUnStrTriangleMesh_element faces_untextured, 'Untextured faces', texture_writer, t_matrix
  end
  
  
  # Get the <vwFaceMesh> strings for the Textured (and Coloured) Faces.
  if !faces_textured.empty?
    
    i=-1  # Counter only used for <vwFaceMesh> name.
    while !faces_textured.empty?
      
      # Update counter and initialise the faces_in_mesh array.
      i+=1
      faces_in_mesh = []
      
      # Grab the Faces in the Faces-with-Texture array that share the same Material.  
      #
      # (This process is done by popping the first element of faces_texture, storing its Material as a 
      # reference, then iteratively popping the [new] first element of faces_texture until the Material of 
      # the first element of faces_texture does not match the "reference" Material.  All popped Faces are 
      # pushed into faces_in_mesh.  This method guarantees that faces_in_mesh will contain all the Faces 
      # in entities [the param to this method] that have the same Material, since faces_texture was sorted 
      # according to Material.)
      #
      # - Note: this can't be performed using the Texture [which would be more direct], since the <=>
      # operator is not defined for Textures [only Materials].)
      common_material = (faces_textured.first)[1]
      while((!faces_textured.empty?) && (common_material == (faces_textured.first)[1]))
        faces_in_mesh.push((faces_textured.delete_at(0))[0])
      end
      
      
      # Call write_faceMesh on faces_in_model and append the result to the textured-Faces vwFaceMesh string.
      model_texturedFaces_faceMesh_string += generate_vwUnStrTriangleMesh_element faces_in_mesh, 'TexturedFaces' + i.to_s, texture_writer, t_matrix
      
    end
    
  end

  # <vwPartActor> element
  actor_body = ''
  actor_body += vws_writer.generate_element('object_name', 'YES', component_name)
  species_type = ''
  if vw_component_type == VW_COMPONENT_TYPE_BUILDING
    species_type = 'ACTOR_BUILDING_PART'
  elsif vw_component_type == VW_COMPONENT_TYPE_CANOPY
    species_type = 'ACTOR_CANOPY_PART'
  elsif vw_component_type == VW_COMPONENT_TYPE_TERRAIN
    species_type = 'ACTOR_TERRAIN_PART'
  end
  actor_body += vws_writer.generate_element('species', 'NO', species_type)
  actor_body += vws_writer.generate_element('wireframe', 'YES', 'OFF')
  actor_body += vws_writer.generate_element('shade', 'YES', 'ON')
  actor_body += vws_writer.generate_element('texture', 'YES', 'ON')
  actor_body += vws_writer.generate_element('skp_color', 'YES', 'OFF')
  actor_body += vws_writer.generate_element('color', 'YES', '255 255 255')
  actor_body += vws_writer.generate_element('xray_intensity', 'YES', '0.0')
  actor_body += vws_writer.generate_element('type', 'NO', 'PHYSICAL')
  actor_body += vws_writer.generate_element('glyph_size', 'NO', '1.0')
  actor_body += vws_writer.generate_element('view_status', 'YES', 'ON')
  actor_body += vws_writer.generate_element('model_status', 'YES', 'ON')
  
  
  # COMPILE THE FINAL RETURN STRING
  
  # Declare the return string and initialise with the string generated from the nested Entities.
  vwComponent_string = nested_components_string
  
  # The top-level String-generation should only occur if there were top-level Faces.
  if((model_untexturedFaces_faceMesh_string != '') || (model_texturedFaces_faceMesh_string != ''))
    
    # Determine the type of vwComponent being generated
    if vw_component_type == VW_COMPONENT_TYPE_BUILDING
      component_type_label = 'vwBuildingPart'
    elsif vw_component_type == VW_COMPONENT_TYPE_CANOPY
      component_type_label = 'vwCanopyPart'
    elsif vw_component_type == VW_COMPONENT_TYPE_TERRAIN
      component_type_label = 'vwTerrainPart'
    else
      component_type_label = 'vwPart'
    end
    
    # Open the <vw[Type]Part> element
    vwComponent_string += vws_writer.generate_open_tag(component_type_label, nil)
    vwComponent_string += vws_writer.generate_element('object_name', 'YES', component_name)
    
    
    # Write the <vwUnStrFaceMeshBundle> element.
    mesh_body = vws_writer.generate_element('object_name', 'YES', component_name + ' mesh bundle')
    mesh_body += model_untexturedFaces_faceMesh_string + model_texturedFaces_faceMesh_string
    
    # Write the <vwPartModel> element (based on the <vwUnStrFaceMeshBundle>).
    model_body = vws_writer.generate_element('object_name', 'YES', component_name + ' model')
    model_body = vws_writer.generate_element('model_status', 'YES', 'ON')
    model_body += vws_writer.generate_element('vwUnStrFaceMeshBundle', nil, mesh_body)
    vwComponent_string += vws_writer.generate_element('vwPartModel', nil, model_body)
    
    # Write the <vwPartActor> element
    vwComponent_string += vws_writer.generate_element('vwPartActor', nil, actor_body)
    
    # Close the <vw[Type]Part> element
    vwComponent_string += vws_writer.generate_close_tag      # Close <vw[...]Part>
    
  end
  
  
  # Return
  return vwComponent_string
  
end



# Helper function to generate a VWS "vwUnStrTriangleMesh" element
# Params:
#     - faces:  An array containing the SketchUp Face objects in the TriangleMesh.
#     - mesh_name:  The name of the Mesh being generated.  If none, pass "" or nil.
#     - texture_writer:  The TextureWriter object which has been loaded with the Textures.  A
#	value of nil indicates that Textures should not be exporter.
#     - t_matrix:   The Transformation object that will be applied to each Entity in entities.
#       (pass identity if no transformation is desired).
# Returns:
#     - The string containing the <vwUnStrTriangleMesh> element.
# Comments: 
#     - This function implicitly assumes the order of Mesh-related traversals (e.g.
#       meshes.each{|mesh|}, mesh.polygons.each{|polygon|}) is identical on every traversal.  This
#       traversal behaviour is not documented anywhere, but has been shown to hold for every 
#       known trial.  Nonetheless, if this assumption is ever proven to be false, this function
#       will have to be written.
#     - On the whole, this function is more complex than what might at first seem necessary,
#       especially with regards to indexing and index variables.  This complexity arises out of
#       the fact that Virtualwind Face Meshes generally contain more than one SketchUp Face.
#       Indexing must therefore be tracked over multiple SketchUp Faces, which is less
#       straightforward than using SketchUp's naive (but per-Face) indexing.
#     - A quick look at the implementation of this function will reveal some rather obvious 
#       inefficiencies.  In particular, loops are often run over the same collections repeatedly,
#       whereas these collections could often have been looped over just once, with some clever
#       optimisations.  These clever optimisations arise by exploiting the tight coupling among 
#       various sub-elements of a <vwUnStrTriangleMesh> (e.g. Triangles are tightly coupled 
#       with Vertices, Colors are tightly coupled with Triangles, and Textures are tightly 
#       coupled with both Vertices and Triangles).  As such, the resulting code is hard to read 
#       and harder to maintain.  These coding ineffeciencies therefore arose out of a 
#       conscious decision to favour readability and modularity over optimisation.  In any event,
#       quick tests suggest that the performance benefits gained by implementing such 
#       optimisations are negligible.
#     - For more information, consult the technical reports pertaining to the SketchUp plugin.
#       and the VWS file-format documentation (esp. the <vwUnStrTriangleMesh> element).
def generate_vwUnStrTriangleMesh_element faces, mesh_name, texture_writer, t_matrix
  
  # INITIALISATION AND PREPROCESSING
  
  # Create the VWSWriter
  vws_writer = VWSWriter.new

  # Return string and component sub-strings
  return_string = ''
  name_string = ''
  vertices_string = ''
  triangles_string = ''
  color_string = ''
  texture_string = ''
  
  
  # Get the (transformed) SketchUp Mesh objects for each Face.
  skupMeshes = []
  faces.each{|face|
    
    skupMesh = face.mesh 7
    skupMesh = skupMesh.transform! t_matrix
    skupMeshes.push(skupMesh)
    
  }



  # CONSTRUCT THE <name> ELEMENT
  
  # Simply construct the element.
  name_string = vws_writer.generate_element('object_name', 'NO', mesh_name)




  # CONSTRUCT THE <vwUnStrVertices> ELEMENT

  # Iterate through the meshes (by index) to populate the vertices.
  vwVertices = []
  skup_to_VW_vertexIndex_map = []
  skupMeshes.each_index{|skupFace_index|
    
    # Get the mesh at the position we're at, and add a row in the 
    # sketchUp_to_VW_index_map.
    skupMesh = skupMeshes.at(skupFace_index)
    skup_to_VW_vertexIndex_map[skupFace_index] = []
    
    # Add each Vertex of the Mesh to the vwVertices array
    skupMesh.points.each_index{|index|
      
      # As a convenience, save index + 1.
      # (This is useful because all Mesh methods are 1-indexed.)
      skupVertex_index = index+1
      
      # Get the actual SketchUp Vertex.
      skupVertex = skupMesh.point_at skupVertex_index
      
      # Put the Vertex in the vwVertices array if it's not already there.  If it is already
      # there, note the index that it's at in this array.  In either case, map this index against
      # the SketchUp Mesh number and SketchUp Vertex number.
      vwVertex_index = vwVertices.index(skupVertex)
      if vwVertex_index != nil
        skup_to_VW_vertexIndex_map[skupFace_index][skupVertex_index] = vwVertex_index
      else
        vwVertices.push skupVertex
        skup_to_VW_vertexIndex_map[skupFace_index][skupVertex_index] = vwVertices.length-1
      end
      
    }
  }

  # Run through the vwVertices array to write out the content of the <vwUnStrVertices> element.
  verticesBody_string = ''
  vwVertices.each_index{|vwVertex_index|
    vertexAttributes_string = "id=\"" + vwVertex_index.to_s + "\""
    vertexBody_string = vwVertices.at(vwVertex_index).to_a.join(' ') + "\n"
    vertexElement = vws_writer.generate_element_with_attributes('vwUnStrVertex', vertexAttributes_string, vertexBody_string)
    verticesBody_string += vertexElement + "\n"
  }

  # Construct the actual <vwUnStrVertices> element.
  vertices_string = vws_writer.generate_element('vwUnStrVertices', nil, verticesBody_string)
  
  
  # CONSTRUCT THE <vwUnStrTriangles> ELEMENT
  
  # Iterate through the Meshes (by index) to populate the triangles
  vwTriangles = []
  skup_to_VW_triangleIndex_map = []
  skupMeshes.each_index{|skupFace_index|
    
    # Add a row in the skup_to_VW_triangleIndex_map.
    skup_to_VW_triangleIndex_map[skupFace_index] = []
    
    # Get the Mesh at the position we're at and get the (SketchUp) 
    # Triangles for the mesh.
    skupMesh = skupMeshes.at(skupFace_index)
    skupTriangles = skupMesh.polygons
    
    # Store each (SkUp) Triangle in vwTriangles
    skupTriangles.each_index{|skupTriangle_index|
      
      # Get the actual SkUp Triangle
      skupTriangle = skupTriangles.at(skupTriangle_index)
      
      # Convert the Vertices of each Triangle from SkUp Vertices to VW Vertices
      vwTriangle = []
      skupTriangle.each{|skupVertex_index|
        # Look up the skupVertex_index in the skup-to-VW index map.
        # (Note: the index must be looked up by absolute value, since SkUp
        # uses the sign bit to encode additional (and irrelevent, for these 
        # purposes) information.
        vwTriangle.push(skup_to_VW_vertexIndex_map[skupFace_index][skupVertex_index.abs])
      }
      
      # Put the (VW) Triangle into the vwTriangles array, and map the index against
      # its skupFace_index and skupTriangle_index.
      vwTriangles.push(vwTriangle)
      skup_to_VW_triangleIndex_map[skupFace_index][skupTriangle_index] = vwTriangles.length-1
    }
    
  }

  # Run through the vwTriangles array to write out the content of the <vwUnStrTriangles> element.
  trianglesBody_string = ''
  vwTriangles.each_index{|vwTriangle_index|
    triangleAttributes_string = "id=\"" + vwTriangle_index.to_s + "\""
    triangleBody_string = vwTriangles.at(vwTriangle_index).to_a.join(' ') + "\n"
    triangleElement = vws_writer.generate_element_with_attributes('vwUnStrTriangle', triangleAttributes_string, triangleBody_string)
    trianglesBody_string += triangleElement + "\n"
  }

  # Construct the actual <vwUnStrTriangles> element.
  triangles_string = vws_writer.generate_element('vwUnStrTriangles', nil, trianglesBody_string)


  
  # CONSTRUCT THE <Color> ELEMENT
  
  # Iterate through the Face (by index) to generate the Color information.
  # We'll iterate through the Faces (not the Meshes, directly), since Color
  # information is contained in the Faces.
  color_body = ''
  rgb_body = ''
  prev_material = nil
  faces.each_index{|skupFace_index|
    
    # Get the Face's Material.
    face_material = faces.at(skupFace_index).material
    
    # Only do something in the Face has a Material.
    if(face_material != nil)
      
      # Get the Color of the Face.
      face_color = face_material.color
      
      # If the current Color is different from the previous Color, that
      # means we have to close the current <vwRGBValue> tag (if one 
      # exists) and set up a new one.
      if(face_material != prev_material)
        if rgb_body != ''
          color_body += vws_writer.generate_element('vwRGBValues', nil, rgb_body)
        end
        rgb_body = face_color.to_a[0..2].join(' ')
        prev_material = face_material
      end
      
      # Append the (VW) Triangles associated the current Face (by looking
      # them up against the skup_to_VW_triangleIndex_map).
      rgb_body += ' ' + skup_to_VW_triangleIndex_map[skupFace_index].join(' ')
      
    end
    
  }
  
  # Append the last rgb_body, if there is one.
  if rgb_body != ''
    color_body += vws_writer.generate_element('vwRGBValues', nil, rgb_body)
  end
  
  # Wrap up the color_body string (if it exists) in <Color> tags store in the color_string
  if color_body != ''
    color_string = vws_writer.generate_element('Color', nil, color_body)
  end
  
  

  # CONSTRUCT THE <Texture> ELEMENT
  
  # Get the Material of the first Face.  (We only have to check the first Face
  # since a TriangleMesh can have at most one Texture.)
  first_face = faces.first
  material = first_face.material
  
  # Only do anything if there actually is a Material, and if the 
  # texture_writer is non-nil.
  if((texture_writer != nil) && (material != nil))
    
    # Get the Texture
    texture = material.texture
    
    # Only continue if there actually is a Texture
    if texture != nil
      
      # Get the filename by using the "Texture Writer Handler" for the first Face.
      # (It's sufficient to check the first Face, since we know these Faces share 
      # the share Texture image, by definition.)
      uvHelper = first_face.get_UVHelper(true, true, texture_writer)
      handle = texture_writer.handle(first_face, true)
      filename_attribute_string = "filename=\"" + (texture_writer.filename handle) + "\""
      
      # Perform the actual Texture-information extraction and storage.  
      # Recall that Texture Coordinates in a VWS file are stored on a 
      # triangle-by-triangle basis, not a vertex-by-vertex basis (as in SkUp).
      # This section of code therefore looks up each Texture Coordinate of 
      # each SkUp Vertex of each SkUp Triangle of each SkUp Mesh.  
      vwTCoords = []
      skupMeshes.each_index{|skupMesh_index|
        
        # Get the Mesh Itself and the Face that it's a Mesh of.
        skupMesh = skupMeshes.at(skupMesh_index)
        face = faces.at(skupMesh_index)
        
        # Get the UV Helper and the "Texture Writer Handle" for the Face.
        uv_helper = face.get_UVHelper(true, true, texture_writer)
        handle = texture_writer.handle(face, true)
        
        # Iterate over each polygon in the Mesh.
        skupMesh.polygons.each{|triangle|
          
          vwTCoords_forTriangle = []
          triangle.each{|vertex|
            # Get the position of each Vertex, then pass that position to the UVHelper
            # to get the UV coord of that point.  (I don't fully understand why, but you
            # have to apply the inverse of the Transformation Matrix to this position in
            # order for the Textures to be mapped correctly when the geometry is in a
            # nested group.)
            position = skupMesh.point_at(vertex)
            position = position.transform(t_matrix.inverse)  
            vwTCoords_forTriangle.push(uv_helper.get_front_UVQ(position))
          }
          vwTCoords.push(vwTCoords_forTriangle)
        }
        
      }
      
      # Run through the vwTCoords array to write out the <vwTextureCoords> 
      # elements.
      vwTCoords.each_index{|vwTCoord_index|
        tcoordAttributes_string = "id=\"" + vwTCoord_index.to_s + "\""
        vwTCoords_forTriangle = vwTCoords.at(vwTCoord_index)
        textureCoords_body = ''
        vwTCoords_forTriangle.each{|vwTCoord|
          textureCoords_body += vwTCoord.to_a[0..1].join(' ')
          textureCoords_body += ' '
        }
        texture_string += vws_writer.generate_element_with_attributes('vwTextureCoords', tcoordAttributes_string , textureCoords_body)
      }
      
      
      # Construct the actual <Texture> element by wrapping the 
      # texture_string in <Texture> tags (and adding the filename attribute)
      texture_string = vws_writer.generate_element_with_attributes('Texture', filename_attribute_string, texture_string)
      
    end
    
  end
  
  

  # RETURN STRING GENERATION
  
  
  # Assemble the return string body
  return_string += name_string
  return_string += vertices_string
  return_string += triangles_string
  return_string += color_string
  return_string += texture_string
  
  # Wrap the return string body in <vwUnStrTriangleMesh> tags and return
  return_string = vws_writer.generate_element('vwUnStrTriangleMesh', nil, return_string)
  return return_string

end



class VWSWriter

  # initialize method
  # Invoked automatically by calls to VWSWriter.new
  def initialize
    
    # Initialise the Tag Stack.  
    # This stack holds the currently open tag names.
    @tag_stack = []
    
  end
  
  # Helper method to write vws elements of the form 
  # <[tag_name] editable="[editable_attribute]">
  # Pushes tag_name onto the @tag_stack.
  # Params:
  #   - tag_name: the name of the element to create.  Is pushed onto @tag_stack.
  #   - editable_attribute: the value to store in the "editable" attribute of the element.
  #     A value of '' or nil will cause the editable attribute to be omitted entirely.
  # Returns: 
  #   a string of the form <[tag_name] editable="[editable_attribute]">;  the editable 
  #   attribute is omitted if editable_attribute parameter is nil or ''.
  def generate_open_tag(tag_name, editable_attribute)
    
    # Construct the tag_string.
    tag_string = '<' + tag_name
    tag_string += (" editable=\"" + editable_attribute + "\"") if(editable_attribute != nil && editable_attribute != '')
    tag_string += ">\n"
    
    # Push the tag_string onto the Tag Stack and return.
    @tag_stack.push tag_name
    return tag_string
    
  end


  # Helper method to write vws elements of the form </[tag_name]>, where tag_name is 
  # retrieved by popping the @tag_stack.
  # Returns: 
  #   a string of the form </[tag_name]>.
  def generate_close_tag
    
    # Construct the tag_string
    tag_string = "</" + @tag_stack.pop + ">\n"
    
    # Return
    return tag_string
    
  end
  
  
  # Helper method to write complete VWS elements of the form:
  #     <[element_name] editable="[editable_attribute]">
  #     [element_body]
  #     </element_name>
  # If [element_body] is nil or '', the return value has the form:
  #     <[element_name] [...] />
  # Params:
  #   - element_name: the name of the element to create.
  #   - editable_attribute: the value to store in the "editable" attribute of the element.
  #     A value of '' or nil will cause the editable attribute to be omitted entirely.
  #   - element_body: the body of the tag.  May be a bottom-level string, or may
  #     itself be a VWS element.  If nil or '', the return element will have the 
  #     form <[element_name] />.  
  # Returns: 
  #   a string of the form:
  #     <[element_name] editable="[editable_attribute]">
  #     [element_body]
  #     </element_name>
  #   if element_body is present; otherwise, return string has the form
  #     <[element_name] [...] />.  
  #   The editable attribute is omitted if editable_attribute parameter is nil or ''.
  # Comments:
  #   - although the the Tag Stack may change during the execution of this method, 
  #     it will have the same contents on termination as it had on invocation.
  #   - generate_element does not insert an explicit terminating "\n" string after 
  #     [element_body].  For the closing </element_name> to appear on a new line,
  #     [element_body] itself must be "\n"-terminated.
  def generate_element(element_name, editable_attribute, element_body)
    
    # Generate the element_string
    if(element_body != nil && element_body != '')
      # If element_body is not empty, construct a string of the form:
      #     <[element_name] editable="[editable_attribute]">
      #     [element_body]
      #     </element_name>
      element_string = generate_open_tag element_name, editable_attribute
      element_string += element_body
      element_string += generate_close_tag
    else
      # If element_body is empty, construct a string of the form:
      #     <[element_name] editable="[editable_attribute]"/>
      element_string = '<' + element_name
      element_string += (" editable=\"" + editable_attribute + "\"") if(editable_attribute != nil && editable_attribute != '')
      element_string += " />\n"
    end
    
    # Return
    return element_string
    
  end


  # Helper method to write complete VWS elements of the form:
  #     <[element_name] [attribute_string]>
  #     [element_body]
  #     </element_name>
  # If [element_body] is nil or '', the return value has the form:
  #     <[element_name] [attribute_string] />
  # Params:
  #   - element_name: the name of the element to create.
  #   - attribute_string: the attribute string of the element being written.  Note 
  #     that attribute_string will be written out to the element verbatim,
  #     separated from the element name by a single space character.  As an
  #     example, to output the element:
  #       <myTag attrib1="att1" attrib2=a2>
  #     attribute_string must be passed a value of "attrib1=\"att1\" attrib2=a2".
  #     A value of '' or nil will cause the element to have no attributes; in such
  #     cases however, the method generate_element() should be preferred.
  #   - element_body: the body of the tag.  May be a bottom-level string, or may
  #     itself be a VWS element.  If nil or '', the return element will have the 
  #     form <[element_name] />.  
  # Returns: 
  #   a string of the form:
  #     <[element_name] [attribute_string]>
  #     [element_body]
  #     </element_name>
  #   if element_body is present; otherwise, return string has the form
  #     <[element_name] [attribute_string] />
  # Comments:
  #   - although the the Tag Stack may change during the execution of this method, 
  #     it will have the same contents on termination as it had on invocation.
  #   - generate_element does not insert an explicit terminating "\n" string after 
  #     [element_body].  For the closing </element_name> to appear on a new line,
  #     [element_body] itself must be "\n"-terminated.
  def generate_element_with_attributes(element_name, attribute_string, element_body)
    
    # Open the element
    element_string = '<' + element_name + ' ' + attribute_string
    
    # If the element_body is empty/null, close the tag immediately.
    if(element_body == nil || element_body == '')
      element_string += " />"
      
    # Otherwise, add the element_body to the element, then close the tag.
    else
      element_string += ">\n"
      element_string += element_body
      element_string += "</" + element_name + '>'
    end
    
    # Return
    return element_string
    
  end

end



# Declare the file is loaded
file_loaded('exportScript.rb')
