# Plugin that provides the user with utilities to attempt to correct corrupt geometry.
# Author: K. Di Filippo

# Global requires
require 'sketchup.rb'


# Top-level function that removes all Edges having exactly zero connected Faces.
def remove_zero_faced_edges
  
  # Get the Entities in a model
  model = Sketchup.active_model
  entities = model.entities
  
  # Get the Edges in the model that have zero faces.
  zero_faced_edges = []
  entities.each{|entity| zero_faced_edges.concat(find_zero_faced_edges(entity))}
  
  # Get the size of zero_face_edges (will be used for display purposes)
  erased_edges = zero_faced_edges.length
  
  # Delete all Edges in zero_faced_edges 
  # (though take care not to delete the same Entity twice, e.g. in Components).
  model.start_operation 'Remove edges'
  zero_faced_edges.each{|edge| edge.erase! if edge.valid?}
  model.commit_operation
  
  # Display a termination message to the user
  UI.messagebox 'Removed ' + erased_edges.to_s + ' zero-faced Edge(s).' 
  
end


# Worker function that returns an array of all Edges having 
# exactly zero connected Faces.
# Params:
#   - entity: the Entity under consideration.  If Entity is a composite
#     find_zero_faced_edges recurses down.
# Returns: 
#   - an array of all Edges having exactly zero connected Faces.
def find_zero_faced_edges entity

  # Initialise the return array.
  zero_faced_edges = []
  
  # If Entity is a zero-faced Edge, add it to the return array;
  # otherwise, if Entity is a composite, return down for each sub-Entity.
  if entity.kind_of?(Sketchup::Edge) && entity.faces.length == 0 
    zero_faced_edges = [entity]
  elsif entity.is_complex_entity?
    entity = entity.definition
    entity.entities.each{|sub_entity| zero_faced_edges.concat(find_zero_faced_edges(sub_entity)) }
  end
  
  # Return
  return  zero_faced_edges

end


# Declare the file is loaded
file_loaded('geometryCorrectionScripts.rb')