# Plugin that detects corrupt geometry within a SketchUp model or selection.
# Author: K. Di Filippo

# Global requires
require 'sketchup.rb'

# Load dependencies
PLUGIN_PATH = 'Virtualwind/'
Sketchup.load PLUGIN_PATH + 'utilityScripts'



# Top-level function that hides all valid Entities and shows all invalid ones.
# Params:
#   - entity_list: the "list of Entities" to be validated.  Here, "list of Entities" refers to any object
#     that will return an object of type Entity when entity_list.each is called.
def show_invalid_entities entity_list

  # Create and start the Progress Bar.
  progress_bar = vwProgressBar.new
  progress_bar.start
  
  # Separate out the valid Edges from the invalid ones.
  separated_edges = separate_valid_and_invalid_edges entity_list
  valid_edges = separated_edges[0]
  invalid_edges = separated_edges[1]
  
  # If no invalid geometry was found, tell the user.
  if invalid_edges.empty?
    UI::messagebox('No non-Watertight geometry found (' + valid_edges.length.to_s + ' Edges checked).')
    progress_bar.clear_display
    
  # Otherwise, tell the user invalid geometry was found, and ask whether to hide valid geometry.
  else
    result = UI::messagebox('Non-Watertight geometry found (' + invalid_edges.length.to_s + ' Edge(s)).  Show only non-Watertight geometry?', MB_YESNO)
    if result == 6
      
      # Reset the Progress Bar now that we have a rough idea of the operation length.
      progress_bar.reset(valid_edges.length + invalid_edges.length)
      
      # Hide all valid Edges and connected Faces and show all invalid Edges and 
      # connected Faces.  
      # Note:  The Show/Hide order is important.  In particular, since we want all 
      # invalid geometry to be shown, and since any given Face may be connected to 
      # any number of both valid Edges and Invalid Edges, Show must occur after Hide.
      valid_edges.each{ |edge| 
        progress_bar.increment
        edge.connected_hidden=true 
      }
      invalid_edges.each{ |edge| 
        progress_bar.increment
        edge.connected_visible=true 
      }
    else
      
      # If the user chooses to abort, clear the Progress Bar.
      progress_bar.clear_display
      
    end
    
  end
  
end


# Helper function that determines and separates all valid and invalid Edges with an input 
# list of Entities.  separate_valid_and_invalid_entities recurses down all complex Entities 
# in entity_list.
# Params:
#   - entity_list: the "list of Entities" to be examined.  Here, "list of Entities" refers to any 
#     object that will return an object of type Entity when entity_list.each is called.
# Returns:
#   - An array of the form [valid_edges, invalid_edges], where (in)valid_edges is itself a list
#     that contains all (in)valid Edges in entity_list.  separate_valid_and_invalid_entities
#     recurses down all complex Entities in entity_list.
# Comments:
#   - An "Edge" is valid iff it has exactly 2 connected Faces.
def separate_valid_and_invalid_edges entity_list
  
  # Declare and initialise the valid and invalid lists.
  valid_list = []
  invalid_list = []
  
  # Perform the separation.
  entity_list.each{ |entity|
    
    # If the Entity is an Edge, it is valid if it has exactly two Faces,
    # and invalid otherwise.
    if entity.kind_of? Sketchup::Edge
      if entity.faces.length == 2
        valid_list.push entity
      else
        invalid_list.push entity
      end
    
    # If the Entity is a complex Entity, recurse down.
    elsif entity.is_complex_entity?
      
      # Get the Entity definition and recurse down.
      entity = entity.definition
      sub_separation = separate_valid_and_invalid_edges entity.entities
      
      # Append the results of the recursion to this level's results.
      valid_list.concat sub_separation[0]
      invalid_list.concat sub_separation[1]
      
    end
    
  }
  
  # Return
  return [valid_list, invalid_list]
  
end



# Top-level function that validates geometry of an "Entity list" (e.g. an Entities object or a Selection object).
# Params:
#   - entity_list: the "list of Entities" to be validated.  Here, "list of Entities" refers to any object
#     that will return an object of type Entity when entity_list.each is called.
# Returns:
#    - false if validation fails on any Entity in entity_list, or if entity_list is nil.
#    - true otherwise.
def validate_entities entity_list

  # Throw an error if entity_list is nil
  if entity_list == nil
    UI.messagebox('Error: entity_list is nil.')
    return false
  end

  # Throw a warning in case entity_list is a Selection, and if View --> Hidden Geometry is turned off.
  if entity_list.is_a?(Sketchup::Selection) && !Sketchup.active_model.rendering_options['DrawHidden'] 
    warning_message = "Warning: View --> Hidden Geometry is turned off.  Check for Watertightness (Selected) may be unreliable.\n\n"
    warning_message += 'Ok to proceed?'
    result = UI.messagebox(warning_message, MB_YESNO)
    if result == 7
      return false
    end
  end
  
  # Initialise the entity counters
  num_edges = 0
  num_faces = 0
  num_groups = 0
  num_components = 0
  
  # Iterate through all Entities in the selection
  entity_list.each{|entity| 
    
    # Call the validator
    result = validate_entity entity
    
    # If the result is false, validation has failed, so return (with false)
    if result == false
      return result
      
    # If the result is not false, add up the counters accordingly
    else
      num_edges += result[0]
      num_faces += result[1]
      num_groups += result[2]
      num_components += result[3]
    end
  }
  
  # If execution made it here, that means all validation suceeded.  Output a success message.
  success_message = "Check complete: geometry is Watertight.\n"
  success_message = success_message + '(Checked ' + num_edges.to_s + ' Edges, ' + num_faces.to_s + ' Faces, ' + num_groups.to_s + ' Groups, and ' + num_components.to_s + ' Components.)'
  UI.messagebox(success_message)

  # Return
  return true

end


# Function that checks the validity of a specific Entity.  In the event of an invalid Entity,
# validate_entity throws an validation-fail message via UI::messagebox and hides all 
# valid components.
# Params:
#    - entity: the Entity object to validate.
# Returns:
#    - false if entity fails validation.
#    - [num_edges, num_faces, num_groups, num_components], where num_[ABC] is the 
#      number of valid Entity objects of type [ABC] within entity (including entity itself).
#    - Examples: calling validate_entity on a valid Face will return [0,1,0,0]; calling 
#      validate_entity on a valid Component with 42 Edges, 8 Faces, and 7 sub-Groups
#      will return [42, 8, 7, 1]; calling validate_entity on an invalid Entity will always
#      return false, regardless of the type of Entity.
# Notes:
#    - An Edge passes validation if and only if it is used by exactly two Faces.
#    - A Face always passes validation.
#    - A composite Entity (e.g. Group, Component) passes validation if and only if
#      all its children Entity objects pass validation.
def validate_entity entity
  
  # Get the Sketchup objects
  model = Sketchup.active_model
  
  # Initialise the entity counters
  num_edges = 0
  num_faces = 0
  num_groups = 0
  num_components = 0
  
  # Validate the Entity (according to the type of entity)
  if entity.kind_of?(Sketchup::Edge)
    
    # Increment the counter
    num_edges += 1
    
    # Get the faces directly connected to this edge.
    # If there are not exactly two, throw an error and return false.
    num_connected_faces = entity.faces.length
    if num_connected_faces != 2
      
      # Construct the core error message.
      face_string = 'Face'
      face_string << 's' if num_connected_faces != 1
      error_message = 'Non-Watertight geometry detected: Edge has ' + num_connected_faces.to_s + ' connected ' + face_string + ', needs exactly 2.'
      
      # Clear the Selection, hide everything but the invalid Edge and connecting Faces, 
      # and zoom the View to this geometry
      model.selection.clear
      hide_entities model.entities
      entity.connected_visible=true
      hidden_state = model.rendering_options['DrawHidden'] 
      model.rendering_options['DrawHidden']=false
      model.active_view.zoom_extents
      model.rendering_options['DrawHidden']=hidden_state
      
      # Display the messagebox and return.
      UI.messagebox(error_message)
      return false
      
    end
  
  # If the entity is a Face, simply increment the counter
  elsif entity.kind_of?(Sketchup::Face)
    num_faces = num_faces + 1
    
  # If the entity is a Group or Component, recurse down and update the counters as you go.
  elsif entity.is_complex_entity?
    entity = entity.definition
    if entity.kind_of?(Sketchup::Group)
      num_groups = num_groups + 1
    else
      num_components = num_components + 1
    end
    entity.entities.each{ |sub_entity| 
      result = validate_entity sub_entity
      if result == false
        return result
      else
        num_edges = num_edges + result[0]
        num_faces = num_faces + result[1]
        num_groups = num_groups + result[2]
        num_components = num_components + result[3]
      end
    }
    
  end
  
  # Return
  return [num_edges, num_faces, num_groups, num_components]
  
end


# Function that provides the user with validation statistics over a specific "Entity list".
# Params: 
#    - entity_list: the "list of Entities" for which the validation statistics will be displayed.  
#      Here, "list of Entities" refers to any object that will return an object of type Entity 
#      when entity_list.each is called.
def show_validation_statistics entity_list

  # Call the validation computation method
  validation_stats = compute_validation_statistics entity_list
  
  # Display the results
  message = "Watertightness statistics: non-Watertight Entities found / total Entities found\n"
  message += '     - Edges: ' + validation_stats[4].to_s + ' / ' + validation_stats[0].to_s + "\n"
  message += '     - Faces: ' + validation_stats[5].to_s + ' / ' + validation_stats[1].to_s + "\n"
  message += '     - Groups (incl. subgroups): ' + validation_stats[6].to_s + ' / ' + validation_stats[2].to_s + "\n"
  message += '     - Components (incl. subcomponents): ' + validation_stats[7].to_s + ' / ' + validation_stats[3].to_s + "\n"
  UI.messagebox(message)
  
end


# Function that computes the validation statitistics of a specific "Entity list".
# Params: 
#    - entity_list: the "list of Entities" for which the validation statistics will be computed.  
#      Here, "list of Entities" refers to any object that will return an object of type Entity 
#      when entity_list.each is called.
# Returns:
#    - An eight-element array of the form [num_total_edges, num_total_faces, num_total_groups,
#      num_total_components, num_invalid_edges, num_invalid_faces, num_invalid_groups, 
#      num_invalid_components].  The "group" and "component" statistics include nested 
#      groups/components.
# Notes:
#    - An Edge passes validation if and only if it is used by exactly two Faces.
#    - A Face always passes validation.
#    - A composite Entity (e.g. Group, Component) passes validation if and only if
#      all its children Entity objects pass validation.
def compute_validation_statistics entity_list
  
  # Define and initialise the counters.
  num_total_edges = 0
  num_total_faces = 0
  num_total_groups = 0
  num_total_components = 0
  num_invalid_edges = 0
  num_invalid_faces = 0
  num_invalid_groups = 0
  num_invalid_components = 0
  
  # Iterate through entity_list to compute the statistics
  entity_list.each{|entity|
    
    # If the Entity is an Edge, it passes validation iff it is used by exactly two Faces.
    if entity.kind_of?(Sketchup::Edge)
      num_total_edges += 1
      num_invalid_edges += 1 if entity.faces.length != 2
    
    # If the Entity is a Face, it always passes validation.
    elsif entity.kind_of?(Sketchup::Face)
      num_total_faces += 1
    
    # If the Entity is a composite, check its children.
    elsif entity.is_complex_entity?
      
      # Define a validation flag, to indicate whether the current Entity has passed validation
      valid_flag = true
      
      # Get the Definition
      entity = entity.definition
      
      # Compute the statistics for this Group/Component
      sub_statistics = compute_validation_statistics(entity.entities)
      num_total_edges += sub_statistics[0]
      num_total_faces += sub_statistics[1]
      num_total_groups += sub_statistics[2]
      num_total_components += sub_statistics[3]
      # Add the invalid statistics only if validation failed.
      # (We have to do this check anyway since we need to know if this Group/Component as a whole passed validation.)
      if((sub_statistics[4] + sub_statistics[5] + sub_statistics[6] + sub_statistics[7]) > 0)
        valid_flag = false
        num_invalid_edges += sub_statistics[4]
        num_invalid_faces += sub_statistics[5]
        num_invalid_groups += sub_statistics[6]
        num_invalid_components += sub_statistics[7]
      end
      if entity.kind_of?(Sketchup::Group)
        num_total_groups += 1
        num_invalid_groups += 1 if !valid_flag
      else
        num_total_components += 1
        num_invalid_components += 1 if !valid_flag
      end
    end
  }
  
  # Return
  return [num_total_edges, num_total_faces, num_total_groups, num_total_components, num_invalid_edges, num_invalid_faces, num_invalid_groups, num_invalid_components]
  
end


# Declare the file is loaded
file_loaded('geometryValidationScripts.rb')