require 'sketchup.rb'


# Class that implements the VertexSewer Tool.  This Tool draws an Edge 
# between two user-specified Vertex objects.
class VertexSewer
  
  # Called when you select the tool
  # Initialise the class variables.
  def activate
    @vertex1 = nil
    @vertex2 = nil
    @selection = Sketchup.active_model.selection
    @selection.clear
  end

  # Called when I select another tool or quit SketchUp
  def deactivate(view)        
  end

  # Called when I roll middle mouse button?
  def draw(view)
  end

  # Called when I roll middle mouse button?
  # Called when I bring SketchUp into view. Followed directly by draw
  def getExtents               
  end

  # This is called followed directly by onRButtonDown
  def getMenu(menu)
  end

  # Called when I press the Esc key. Followed directly by onKeyDown
  def onCancel(reason, menu)                
  end

  def onKeyDown(char, rpt, flags, view)
  end

  def onKeyUp(char, rpt, flags, view)
  end

  # NOTE: Called after onLButtonDown and onLButtonUp. 
  def onLButtonDoubleClick(flags, x, y, view)
  end

  # Handle the click event.  This is the main function.  It records the Vertex selected
  # and draws the line if two vertices total have been selected.
  def onLButtonDown(flags, x, y, view)
    
    # Store the selected Vertex and select the neighbouring Edges
    # (just for visualisation purposes).
    inputpoint = view.inputpoint x,y
    if @vertex1 == nil 
      @vertex1 = inputpoint.vertex
      @selection.clear
      @selection.add @vertex1.edges
    elsif @vertex2 == nil
      @vertex2 = inputpoint.vertex
      @selection.add @vertex2.edges
    end
    
    # If both Vertex 1 and Vertex 2 have been selected, create the connecting Edge
    # and clear the stored vertices.
    if @vertex1 != nil && @vertex2 != nil
      Sketchup.active_model.entities.add_edges @vertex1, @vertex2
      @vertex1 = nil
      @vertex2 = nil
    end
  end

  def onLButtonUp(flags, x, y, view)
  end


  # Called a lot so I did not put a puts message in
  def onMouseEnter(view)
  end

  # Called a lot so I did not put a puts message in
  def onMouseLeave(view)
  end

  # Called a lot so I did not put a puts message in
  def onMouseMove(flags, x, y, view)           
  end

  # NOTE: Called after onrButtonDown and onRButtonUp.
  def onRButtonDoubleClick(flags, x, y, view)
  end

  def onRButtonDown(flags, x, y, view)
  end

  # Called not only right after a onRButtonDown, but after a onRButtonDoubleClick?
  def onRButtonUp(flags, x, y, view)
  end

  # NOTE: onReturn is followed directly by onKeyDown
  def onReturn(view)
  end

  def onUserText(text, view)
  end

  # Called when I double-click middle mouse button (exits out of orbit)
  def resume(view)
  end

  # Called when I press middle mouse button (goes into orbit)
  def suspend(view)
  end
  
end



# Class that implements the EdgeSewer Tool.  This Tool draws an tessalation
# between two user-specified Edge objects.
class EdgeSewer
  
  # Called when you select the tool
  # Initialise the class variables.
  def activate
    @edge1 = nil
    @edge2 = nil
    @selection = Sketchup.active_model.selection
    @selection.clear
  end

  # Called when I select another tool or quit SketchUp
  def deactivate(view)        
  end

  # Called when I roll middle mouse button?
  def draw(view)
  end

  # Called when I roll middle mouse button?
  # Called when I bring SketchUp into view. Followed directly by draw
  def getExtents               
  end

  # This is called followed directly by onRButtonDown
  def getMenu(menu)
  end

  # Called when I press the Esc key. Followed directly by onKeyDown
  def onCancel(reason, menu)                
  end

  def onKeyDown(char, rpt, flags, view)
  end

  def onKeyUp(char, rpt, flags, view)
  end

  # NOTE: Called after onLButtonDown and onLButtonUp. 
  def onLButtonDoubleClick(flags, x, y, view)
  end


  # Handle the click event.  This is the main function.  It records the Edges selected
  # and sews them together when two have been selected.
   def onLButtonDown(flags, x, y, view)
    
    # Store the selected Edge and add it to the Selection
    # (just for visualisation purposes).
    inputpoint = view.inputpoint x,y
    if @edge1 == nil 
      @edge1 = inputpoint.edge
      @selection.clear
      @selection.add @edge1
    elsif @edge2 == nil
      @edge2 = inputpoint.edge
      @selection.add @edge2
    end
    
    # If two Edges have been selected, sew them together.
    if @edge1 != nil && @edge2 != nil
      
      # Store the vertices for convenience.
      vertices1 = @edge1.vertices
      vertices2 = @edge2.vertices
      
      # Determine the  neighbouring vertices.
      max_parallel = vertices1[0].position.distance(vertices2[0].position) 
      if vertices1[1].position.distance(vertices2[1].position) > max_parallel
        max_parallel = vertices1[1].position.distance(vertices2[1].position) 
      end
      max_crossed = vertices1[0].position.distance(vertices2[1].position) 
      if vertices1[1].position.distance(vertices2[0].position) > max_crossed
        max_crossed = vertices1[1].position.distance(vertices2[0].position) 
      end
      parallel_config_flag = (max_parallel <= max_crossed)
      
      # Create the bounding Edges.
      entities = Sketchup.active_model.active_entities
      edges = [@edge1, @edge2]
      if parallel_config_flag
        edges.concat(entities.add_edges vertices1[0], vertices2[0])
        edges.concat(entities.add_edges vertices1[1], vertices2[1])
      else
        edges.concat(entities.add_edges vertices1[1], vertices2[0])
        edges.concat(entities.add_edges vertices1[0], vertices2[1])
      end
      
      # Create the Faces
      # (If the two Edges are coplanar, just use these two Edges plus 
      # the two bounding Edges to form the Face.  Otherwise, form the tessalation.)
      if(vertices1[0].position.on_plane?(Geom::fit_plane_to_points(vertices1[1].position, vertices2[0].position, vertices2[1].position)))
        entities.add_face edges
      else
        # Figure out the diagonal first (depending on the configuration of the vertices).
        if parallel_config_flag
          if vertices1[1].position.distance(vertices2[0].position) < vertices1[0].position.distance(vertices2[1].position)
            edges.concat(entities.add_edges vertices1[1], vertices2[0])
          else
            edges.concat(entities.add_edges vertices1[0], vertices2[1])
          end
        else
          if vertices1[0].position.distance(vertices2[0].position) < vertices1[1].position.distance(vertices2[1].position)
            edges.concat(entities.add_edges vertices1[0], vertices2[0])
          else
            edges.concat(entities.add_edges vertices1[1], vertices2[1])
          end
        end
        # Create the Edges.
        entities.add_face edges[0], edges[3], edges[4]
        entities.add_face edges[1], edges[2], edges[4]
      end
      
      # Reset the Edge holders.
      @edge1 = nil
      @edge2 = nil
    end
  end

  def onLButtonUp(flags, x, y, view)
  end


  # Called a lot so I did not put a puts message in
  def onMouseEnter(view)
  end

  # Called a lot so I did not put a puts message in
  def onMouseLeave(view)
  end

  # Called a lot so I did not put a puts message in
  def onMouseMove(flags, x, y, view)           
  end

  # NOTE: Called after onrButtonDown and onRButtonUp.
  def onRButtonDoubleClick(flags, x, y, view)
  end

  def onRButtonDown(flags, x, y, view)
  end

  # Called not only right after a onRButtonDown, but after a onRButtonDoubleClick?
  def onRButtonUp(flags, x, y, view)
  end

  # NOTE: onReturn is followed directly by onKeyDown
  def onReturn(view)
  end

  def onUserText(text, view)
  end

  # Called when I double-click middle mouse button (exits out of orbit)
  def resume(view)
  end

  # Called when I press middle mouse button (goes into orbit)
  def suspend(view)
  end
  
end
