# Scripts relating to texture manipulation
# Author: K. Di Filippo

# Global requires
require 'sketchup.rb'

# Load dependencies
PLUGIN_PATH = 'Virtualwind/'
Sketchup.load PLUGIN_PATH + 'utilityScripts'

# Constants
TEXTURE_PROBLEM_TYPE_NONE = 0
TEXTURE_PROBLEM_TYPE_ANY_PROBLEM = 1
TEXTURE_PROBLEM_TYPE_TWO_DIFFERENT_TEXTURES = 2
TEXTURE_PROBLEM_TYPE_BACK_TEXTURE_WITHOUT_FRONT_TEXTURE = 3

# Top-level function that detects problematic textures in a model.
# Params:
#   - problem_type: the type of problem to flag.  May be any
#     TEXTURE_PROBLEM_TYPE_[...] constant.
def detect_problematic_textures problem_type
  
  # Create and start the Progress Bar.
  progress_bar = vwProgressBar.new
  progress_bar.start
  
  # Get the Entities and set up the "problem" stack.
  model = Sketchup.active_model
  entities = model.entities
  problem_faces = []
  
  # Perform the validation
  entities.each{|entity|
    # (Don't bother to perform the validation on Edges.)
    if !entity.kind_of? Sketchup::Edge
      problem_faces.concat(find_faces_with_problematic_texture entity, problem_type)
    end
  }
  
  # If there were problems, visualise them if the user so chooses.
  if problem_faces.empty?
    UI.messagebox('No problematic textures found.')
    progress_bar.clear_display
  else
    result = UI.messagebox('Problematic textures found.  Show all problematic Faces?', MB_YESNO)
    if result == 6
      
      # Reset the Progress Bar now that we have a rough idea of the operation length.
      progress_bar.reset(problem_faces.length)
      
      # "Start" the action for "Undo" purposes
      model.start_operation 'Hide problem textures'
      
      # Update the visibilities
      hide_entities entities
      problem_faces.each{|face| 
        progress_bar.increment
        face.connected_visible=true
      }
      
      # "Commit" the action for "Undo" purposes
      model.commit_operation
    else
      
      # If the user chooses to abort, clear the Progress Bar.
      progress_bar.clear_display
      
    end
  end
  
end


# Top-level function that detects all valid textures in a model.
def detect_valid_textures
  
  # Create and start the Progress Bar.
  progress_bar = vwProgressBar.new
  progress_bar.start
  
  # Get the Entities and set up the "valid" stack.
  model = Sketchup.active_model
  entities = model.entities
  valid_faces = []
  
  # Perform the validation
  entities.each{|entity| 
    # (Don't bother to perform the validation on Edges.)
    if !entity.kind_of? Sketchup::Edge
      valid_faces.concat(find_faces_with_problematic_texture entity, TEXTURE_PROBLEM_TYPE_NONE)
    end
  }
  
  # If there were problems, visualise them if the user so chooses.
  if valid_faces.empty?
    UI.messagebox('No valid textures found.')
  else
    result = UI.messagebox('Show only validly textured Faces?', MB_YESNO)
    if result == 6
      
      # Reset the Progress Bar now that we have a rough idea of the operation length.
      progress_bar.reset(valid_faces.length)
      
      # "Start" the action for "Undo" purposes
      model.start_operation 'Hide problem textures'
      
      # Update the visibilities
      hide_entities entities
      valid_faces.each{|face|
        progress_bar.increment
        face.connected_visible=true
      }
      
      # "Commit" the action for "Undo" purposes
      model.commit_operation
      
    else
      
      # If the user chooses to abort, clear the Progress Bar.
      progress_bar.clear_display
      
    end
  end
  
end


# Core function to find problematic textures in a model.
# Params:
#   - entity: the entity to check; if entity is complex, this function
#     will recurse down.
#   - problem_type: the problem type fo flag.  May be any
#     TEXTURE_PROBLEM_TYPE_[...] constant.
# Returns:
#   The list of Faces that match the problem_type.  Note that if 
#   problem_type is "NONE", then the return is actually all 
#   *valid* Faces.
def find_faces_with_problematic_texture entity, problem_type
  
  # Declare and initialise the matching_faces stack
  matching_faces = []
  
  # If the Entity is a Face...
  if entity.kind_of?(Sketchup::Face)
    
    # ... determine if the Face has a problem depending on the problem_type parameter.
    matching_flag = false
    if entity.back_material != nil        # (If there's no back_material, then there's never a problem.)
      if problem_type == TEXTURE_PROBLEM_TYPE_ANY_PROBLEM
        matching_flag = true
      elsif problem_type == TEXTURE_PROBLEM_TYPE_TWO_DIFFERENT_TEXTURES
        matching_flag = true if(entity.material != nil && entity.material != entity.back_material)
      elsif problem_type == TEXTURE_PROBLEM_TYPE_BACK_TEXTURE_WITHOUT_FRONT_TEXTURE
        matching_flag = true if(entity.material == nil)
      end
    else
      matching_flag = true if problem_type == TEXTURE_PROBLEM_TYPE_NONE
    end
    
    # If there is a problem, then store the Face into the matching stack.
    matching_faces.push(entity) if matching_flag
    
  # If the Entity is a composite object....
  elsif entity.is_complex_entity?
    # ... recurse down 
    entity = entity.definition
    entity.entities.each{|sub_entity|
      matching_faces.concat(find_faces_with_problematic_texture sub_entity, problem_type)
    }
  end
  
  # Return
  return matching_faces
  
end

# Function that exports all Textures in a model to disk.
# Params:
#   - path: the path where the Textures will be written.
# Returns:
#   - The Texture Writer created and used by the function to store the Textures in.
def write_all_textures_in_model path

  # Return variable
  textures_in_model = []
  
  # Create the TextureWriter
  texture_writer = Sketchup.create_texture_writer
  
  # Write the textures for each Entity
  entities = Sketchup.active_model.entities
  entities.each{|entity| textures_in_model.concat(load_texture_for_entity entity, texture_writer)}
  write_status = texture_writer.write_all (path,  false)
  
  # Return (ensuring that elements are unique)
  return texture_writer
  
end


# Core function that loads a texture_writer with all Textures in an entity
# Params:
#   - entity: the Entity for which the Textures will be loaded.
#   - texture_writer: the TextureWriter object into which the Textures will be loaded.
# Returns:
#   - An array of all texture filenames that were output, according to texture.filename.
#     The elements in this array are guaranteed to be unique.
def load_texture_for_entity entity, texture_writer

  # Return variable
  textures_in_entity = []
  
  # If the Entity is a Face, load the front and back Textures (if present)
  if entity.kind_of? Sketchup::Face
    material = entity.material
    if material != nil
      texture = material.texture
      if texture != nil
        texture_writer.load entity, true
        textures_in_entity.push(texture.filename)
      end
    end
    back_material = entity.back_material
    if back_material != nil
      back_texture = back_material.texture
      if back_texture != nil
        texture_writer.load entity, false
        textures_in_entity.push(back_texture.filename)
      end
    end
    
  # If the Entity is a composite, recursively load the Textures of its child objects.
  elsif entity.is_complex_entity?
    entity = entity.definition
    entity.entities.each{|sub_entity| textures_in_entity.concat(load_texture_for_entity sub_entity, texture_writer) }
  end
  
  # Return
  return textures_in_entity.uniq
  
end


def reverse_orientation_and_texture_for_backtextured_faces
  
  # Get the Entities and set up the "problem" and "valid" stacks.
  model = Sketchup.active_model
  entities = model.entities
  problem_faces = []
  
  # Perform the problem-check
  problem_type = TEXTURE_PROBLEM_TYPE_BACK_TEXTURE_WITHOUT_FRONT_TEXTURE
  entities.each{|entity| problem_faces.concat(find_faces_with_problematic_texture entity, problem_type)}
  
  # For each problem Face, reverse the orientation and the texture side (if the user accepts).
  if problem_faces.empty?
    UI.messagebox('No Faces with problematic Textures found.')
  else
    result = UI.messagebox('Found ' + problem_faces.length.to_s + ' Face(s) with back Texture but no front Texture.  Flip these faces?', MB_YESNO)
    if result == 6
      problem_faces.each{|face| 
        face.reverse!
        face.material = face.back_material
        face.back_material = nil
      }
      # Confirm completion and hide all non-flipped Faces (if user accepts).
      result = UI.messagebox('Back-textured Faces have been flipped.  Show these Faces?', MB_YESNO)
      if result == 6
        model.start_operation 'Hide flipped textures'
        hide_entities entities
        problem_faces.each{|face| face.connected_visible=true}
        model.commit_operation
      end
    end
  end

end


# Declare that the script file is loaded
file_loaded('textureScripts.rb')
