# General-purpose "utility" scripts
# Author: K. Di Filippo

# Global requires
require 'sketchup.rb'

# Load dependencies
PLUGIN_PATH = 'Virtualwind/'
Sketchup.load PLUGIN_PATH + 'vwGlobalConstants'


# Top-level function that classifies each Entity in a list as either a "Solid" or a "Sheet",
# according to the Virtualwind concepts.  The classification is performed using SketchUp's
# Layers: solids and sheets are assigned to two separate Layers.  A complex Entity is 
# classified as a solid if all of its sub-Entities are solids; otherwise, if at least one 
# sub-Entity is a sheet, the complex Entity is classified as a sheet.
# Params:
#   - entity_list: The list of Entity objects to categorise.  entity_list.to_a must 
#     return an array of Entity objects.
# Returns:
#   - true if all entities in entity_list are solids; false otherwise.
def separate_solids_and_sheets entity_list
  
  # Initialise the solid_flag variable (which will be the return variable too)
  solid_flag = true
  
  # Define the new layers (or grab the existing layers, if they exist).
  layers = Sketchup.active_model.layers
  solidLayer = layers.add LAYER_NAME_SOLIDS
  sheetLayer = layers.add LAYER_NAME_SHEETS
  imageLayer = layers.add LAYER_NAME_IMAGES
  unhandledLayer = layers.add LAYER_NAME_UNHANDLED

  # Make sure entity_list is an array
  entity_list = entity_list.to_a
  
  # Create and start the Progress Bar.
  original_length = entity_list.length
  progress_bar = vwProgressBar.new(original_length)
  progress_bar.start

  # Separate the Solids from the Sheets.
  while !entity_list.empty?
    
    # Get the "seed" Entity by popping from the entity_list.
    seed_entity = entity_list.pop
    
    # If the seed Entity is a complex Entity...
    if seed_entity.is_complex_entity?
      
      # Make sure that the Entity isn't a Terrain or an Image.
      if seed_entity.layer.name != LAYER_NAME_TERRAIN
        
        #... Get the definition of the seed Entity
        # (We can't overwrite the Instance the way we normally do because we'll need to operate on it.)
        seed_entity_definition = seed_entity.definition
        
        # Recurse down on the seed Entity's Entity list and categorise the seed Entity
        # based on the result of the recursion.
        sub_group_flag = separate_solids_and_sheets seed_entity_definition.entities
        if sub_group_flag
          seed_entity.layer = solidLayer
        else
          seed_entity.layer = sheetLayer
        end
        solid_flag &= sub_group_flag
        
      end
    
    
    # If the seed Entity is a Face or Edge...
    elsif((seed_entity.kind_of? Sketchup::Face) || (seed_entity.kind_of? Sketchup::Edge))
      
      # ... Get all Entities connected to the seed Entity and delete them from 
      # the entity_list array.
      all_connected_entities = seed_entity.all_connected
      all_connected_entities.each{|entity| entity_list.delete(entity)}
      
      # Perform the separation.
      # (If all the connected Entities pass validation, categorise all these Entities as a Solid;
      # otherwise, categorise them as a Sheet and turn off the solid_flag.)
      if (validate_edges all_connected_entities)
        all_connected_entities.each{|entity| entity.layer=solidLayer}
      else
        all_connected_entities.each{|entity| entity.layer=sheetLayer}
        solid_flag = false
      end
      
      # Update the Progress Bar
      progress_bar.update(original_length - entity_list.length)
      

    # If the seed Entity is an Image....
    elsif seed_entity.kind_of? Sketchup::Image
      
      # Automatically put it on the Image Layer and move on.
      seed_entity.layer = imageLayer
    

    # If the seed Entity is anything else....
    else
	
	# Automatically put it on the Unhandled Layer and move on.
      seed_entity.layer = unhandledLayer


    end

  end
  
  # Return
  return solid_flag

end


# Top-level function to make the Selection a Terrain (by moving it to the Terrain Layer).
# If make_selection_into_terrain determines that the current Selection is not a valid 
# Terrain, or the operation encounters another problem, the operation will abort.
# Returns:
#   - false if the move did not occur (for whatever reason); otherwise, returns
#     the Entity that was made into a Terrain.
# Notes:
#   - Sketchup.selection must have exactly one element, and it must be a complex Entity.
#     If this is not the case, make_selection_into_terrain will throw an error MessageBox 
#     and abort.
#   - If the entity is already on the Teerrain Layer, make_selection_into_terrain throws 
#     an error MessageBox and aborts.
#   - If the Terrain Layer already exists, make_selection_into_terrain will move any 
#     existing geometry on that layer onto the default Layer, after confirming with the user
#     that this is ok.  If the user disagrees, make_selection_into_terrain aborts.
def make_selection_into_terrain
  
  # Get the Sketchup objects
  model = Sketchup.active_model
  selection = model.selection
  layers = model.layers
  entities = model.entities
  
  # Try to add the Terrain Layer (and determine if it already existed)
  orig_num_layers = layers.count
  terrain_layer = layers.add LAYER_NAME_TERRAIN
  terrainLayer_already_exists = orig_num_layers == layers.count
  
  # --- Error checking ---
  
  # Make sure the Selection has only exactly one Entity.  
  # If it does, get it and continue, otherwise abort.
  if selection.count != 1
    error_message = 'Please select exactly one top-level Group or Component to make into a Terrain.'
    UI.messagebox error_message
    return false
  else
    terrain_entity = selection.first
  end
  
  # If the terrain_entity is already on the Terrain Layer, tell the user and exit.
  if terrain_entity.layer == terrain_layer
    UI.messagebox('Entity is already a Terrain.')
    return false
  end
  
  # Make sure terrain_entity is a top-level complex Entity.  (Only top-level complex Entities
  # can be assigned to the terrain layer.)
  if !terrain_entity.is_complex_entity? || (!terrain_entity.parent.kind_of? Sketchup::Model)
    error_message = 'Error: the Entity you are trying to designate as Terrain is not a top-level Group or Component.' 
    error_message += '  Only top-level Groups or Components can be assigned as Terrain.'
    UI.messagebox error_message
    return false
  end

  # Make sure the terrain_entity is fully connected.
  # To do this, grab the first sub-Entity and call its connected() method.
  # If terrain_entities is fully connected, the size of terrain_entities
  # will be the same as the size of terrain_entities[0].connected.
  # (Any sub-Entity will work [not only the first one], but we'll pick the
  # first one for definitiveness.)
  # Note that this will automatically detect sub-Group/-Components as
  # well, since Group/Components cannot be connected to anything.
  terrain_entities = terrain_entity.definition.entities
  if terrain_entities.at(0).is_complex_entity? || (terrain_entities.at(0).all_connected.length != terrain_entities.count)
    error_message = 'Error: the Entity you are trying to designate as Terrain is not fully-connected'
    error_message += ' or has nested Groups/Components.'
    error_message += '  Only fully-connected geometry without sub-Groups/-Components can be assigned as Terrain.'
    error_message += '  Please correct this and try again.'
    UI.messagebox error_message
    return false
  end
  
  # If the Terrain Layer already existed, warn the user that anything already
  # on the Terrain Layer will be placed onto the default Layer.
  if terrainLayer_already_exists
    
    # Prepare and display the warning.
    warning_message = 'Warning: Terrain Layer already exists.'
    warning_message += '  Any Entity currently on the Terrain Layer will be placed on the default layer.'
    warning_message += "\n\nOk to proceed?"
    result = UI.messagebox(warning_message, MB_YESNO)
    if result == 6
      # If the user accepts, perform the actual move
      default_layer = layers[0]
      entities.each{|entity|  set_layer_recursively(entity, default_layer) if entity.layer == terrain_layer }
    else
      # If the user doesn't accept, abort the operation.
      return false
    end
    
  end

  # --- Error checking finished ---
  
  
  # If we get here, that means all error checks and warning checks have passed.
  # All that's left is to actually move the Entity to the Terrain Layer.
  set_layer_recursively terrain_entity, terrain_layer
  return terrain_entity
  
end



# Top-level function that unhides all Entity objects in a model.
# Needed because the "unhide" feature provided by SketchUp does not 
# handle composite objects.
def unhide_all_entities
  
  # Get the Entities
  entities = Sketchup.active_model.entities
  
  # Create and start the Progress Bar.
  progress_bar = vwProgressBar.new(entities.length)
  progress_bar.start

  # Unhide all the Entities.
  entities.each{|entity| 
    unhide_entity entity
    progress_bar.increment
  }
  
end


# Core function that unhides all Entity objects.
# Params:
#   - entity: the Entity object to unhide.  If entity is a composite,
#     unhide_entity recurses down.
def unhide_entity entity
  
  # If the Entity is an Edge or Face, just unhide it
  if entity.kind_of?(Sketchup::Edge) || entity.kind_of?(Sketchup::Face)
    entity.visible=true
  # If the Entity is a complex object, recurse down
  elsif entity.is_complex_entity?
    entity = entity.definition
    entity.entities.each{|sub_entity| unhide_entity sub_entity }
  end
  
end


# Helper function to hide Entity objects.
# Params:
#   - entities: the "list of Entities" to be validated.  Here, "list of Entities" refers to any object
#     that will return an object of type Entity when entities.each is called.
# Comments: 
#   - Intended as a helper function; not intended as a top-level function.
def hide_entities entities
   entities.each{|entity|
    if entity.is_complex_entity?
      hide_entities entity.definition.entities
    else
      entity.hidden=true
    end
   }
end


# Helper function that determines if all the Edge objects in a list of
# Entities are valid.
# Params:
#   - entity_list: The list of Entity objects to validate over.  
#   entity_list.each must produce a list of Entity objects.
# Returns:
#   - false if entity_list contains one or more invalid Edges; true otherwise.
def validate_edges entity_list
  
  entity_list.each{|entity|
    if entity.kind_of? Sketchup::Edge 
      if !entity.valid_solid_edge?
        return false
      end
    end
  }
  
  return true
  
end


# Top-level function that sets the Layer of an Entity and all its children 
# to a specified Layer.
# Params:
#   - entity: the Entity for which to set the Layer.  If Entity is complex, 
#     set_layer_recursively sets the Layer of all its children.
#   - layer: the Layer to which Entity and all its children will be set to.
def set_layer_recursively entity, layer
  
  entity.layer = layer
  if entity.is_complex_entity?
    entity.definition.entities.each{|sub_entity| set_layer_recursively(sub_entity, layer)}
  end
  
end



# Declare that the script file is loaded
file_loaded('utilityScripts.rb')
