# Class definition of the vwProgressBar.
# Author: K. Di Filippo

# Path
PLUGIN_PATH= 'Virtualwind/'

# Global requires
require 'sketchup.rb'
Sketchup.load PLUGIN_PATH + 'vwGlobalConstants'


# Define the vwProgressBar class
class vwProgressBar

  # Initialisation method.  Invoked automatically by calls to vwProgressBar.new
  # Params:
  #   - max_increment: the value that corresponds to a Progress of 100%
  #     Default value: 10.
  #   - num_increments: the number of incrementation "ticks" to be displayed.
  #     Default value: 10.
  def initialize max_increment=10, num_increments=10
    
    # Set up the instance variables.
    set_num_increments num_increments
    set_max_increment max_increment
    @current_increment = 0.0
    
  end

  
  # Method to cause the Progress Bar to display itself initially,
  # and also changes the cursor to a "busy" state.
  def start
    
    Sketchup.set_status_text 'Progress:',  SB_VCB_LABEL
    self.update 0.0
    
    # Change the cursor to the "busy" state.
    cursor_image = Sketchup.find_support_file('Plugins\Virtualwind\icons\cursor_busy.png')
    cursor_id = UI.create_cursor cursor_image, 0, 0
    UI.set_cursor cursor_id
    
  end
  
  
  # Method to reset the current_increment of the Progress Bar, with the option 
  # of specifying new values for the max_increment and the num_increments.
  # Calling this reset method causes the vwProgressBar to restart.
  # Params:
  #   - new_max_increment: the new value for max_increment.  Default value: nil.
  #     If nil, the value of max_increment is left unchanged.
  #   - new_num_increments: the new value for num_increments.  Default value: nil.
  #     If nil, the value of max_increment is left unchanged.
  def reset new_max_increment=nil, new_num_increments=nil
    
    set_max_increment(new_max_increment) if new_max_increment != nil
    set_num_increments(new_num_increments) if new_num_increments != nil
    current_increment = 0.0
    self.start
    
  end
  

  # Method that causes the Progress Bar to increment its current_increment by 1.
  def increment
    
    if @current_increment < @max_increment
      @current_increment += 1 
      self.refresh_display
    end
    
  end


  # Method that updates the Progress Bar to a new (arbitrary) current_increment
  # and refreshes the display.
  # Params:
  #   - current_increment: the new value for this Progress Bar's current_increment.
  def update current_increment
    
    set_current_increment current_increment
    self.refresh_display
    
  end
  
  
  # Method that causes the Progress Bar to update its display.  refresh_display is 
  # called internally every time current_increment is changed.  
  # Note: there is currently no known reason for this function to be called 
  # externally.  It is left public as a courtesy (nothing bad happens if it's called).
  def refresh_display

    if @current_increment >= @max_increment
      status_text = 'Complete'
    else
      num_dashes = (@current_increment.to_f / @max_increment.to_f * @num_increments).floor
      num_dots = @num_increments-num_dashes
      status_text = '|'
      num_dashes.times{status_text += '-'}
      num_dots.times{status_text += '.'}
      status_text += '|'
    end
    
    Sketchup.set_status_text status_text , SB_VCB_VALUE
    
  end


  # Method that clears (i.e. visually removes) the Progress Bar from the screen.
  # Note: clear_display is for visualisation purposes; the internal state of 
  # the vwProgressBar is unaffected.
  def clear_display
    
    Sketchup.set_status_text '', SB_VCB_LABEL
    Sketchup.set_status_text '', SB_VCB_VALUE
    
  end


# "Setter" methods for vwProgressBar's instance variables.  These are declared 
# private since we don't want a user to be able to write directly to a variable.
# They are provided for convenience only, to encapsulate commonly-performed 
# tasks (e.g. validation, sanitizing).

private

  # Sets @num_increments to a new value.
  # Params:
  #   - new_num_increments: the new value for the Progress Bar's 
  #     @num_increments value.
  # Setting rules:
  #   - if new_num_increments > 0, @num_increments <-- new_num_increments;
  #   - otherwise, @num_increments <-- 0.
  def set_num_increments new_num_increments
    if new_num_increments > 0
      @num_increments = new_num_increments
    else
      @num_increments = 0
    end
  end

  # Sets @max_increment to a new value.
  # Params:
  #   - new_max_increment: the new value for the Progress Bar's 
  #     @max_increment value.
  # Setting rules:
  #   - if new_max_increment > 0, @max_increment <-- new_max_increment;
  #   - otherwise, @max_increment <-- 0.
  def set_max_increment new_max_increment
    if new_max_increment > 0
      @max_increment = new_max_increment
    else
      @max_increment = 0
    end
  end
 
  # Sets @current_increment to a new value.
  # Params:
  #   - new_current_increment: the new value for the Progress Bar's 
  #     @current_increment value.
  # Setting rules:
  #   - if new_current_increment < 0, @current_increment <-- 0;
  #   - otherwise if new_current_increment > @max_increment, @current_increment <-- @max_increment;
  #   - otherwise, @current_increment <-- new_current_increment.
  def set_current_increment new_current_increment
    if new_current_increment < 0
      @current_increment = 0.0
    elsif new_current_increment > @max_increment
      @current_increment = @max_increment
    else
      @current_increment = new_current_increment.to_f
    end
  end

end



# Declare the file is loaded
file_loaded('vwProgressBar')
